#
# Copyright (c) 2015-2020 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS_auth_jwt.proxy module

This module defines a proxy handler, which is used when the JWT authentication plug-in
is configured in proxy mode, relying on another authentication authority providing it's
own REST API.
"""

import logging

import requests
from pyramid.httpexceptions import HTTPOk, HTTPUnauthorized

from pyams_auth_jwt.interfaces import IJWTAuthenticationPlugin, IJWTProxyHandler, \
    IJWTSecurityConfiguration, JWT_PROXY_CACHE_NAME, JWT_PROXY_TOKENS_NAMESPACE
from pyams_security.interfaces import ISecurityManager
from pyams_utils.adapter import ContextAdapter, adapter_config
from pyams_utils.cache import get_cache


__docformat__ = 'restructuredtext'


LOGGER = logging.getLogger('PyAMS (JWT security)')


@adapter_config(required=IJWTAuthenticationPlugin,
                provides=IJWTProxyHandler)
@adapter_config(required=ISecurityManager,
                provides=IJWTProxyHandler)
class JWTProxyHandler(ContextAdapter):
    """JWT proxy handler"""

    def get_claims(self, request, obj=None):
        """Get claims from provided request authorization"""
        authorization = getattr(request, 'authorization', None)
        if authorization is None:
            authorization = request.headers.get('Authorization')
        else:
            authorization = ' '.join(authorization)
        if not authorization:
            return HTTPUnauthorized.code, {}
        configuration = IJWTSecurityConfiguration(self.context)

        # Check Beaker cache
        tokens_cache = None
        cache_key = None
        if configuration.use_cache:
            cache_key = authorization
            if obj:
                cache_key = '{}::{}'.format(cache_key, obj)
            tokens_cache = get_cache(JWT_PROXY_CACHE_NAME, configuration.selected_cache,
                                     JWT_PROXY_TOKENS_NAMESPACE)
            try:
                claims = tokens_cache.get_value(cache_key)
            except KeyError:
                pass
            else:
                return HTTPOk.code, claims

        # Call authority REST API
        method, service = configuration.get_claims_service  # pylint: disable=unpacking-non-sequence
        rest_service = '{}{}'.format(configuration.authority, service)
        params = {}
        data = {}
        if obj:
            if method == 'GET':
                params['obj'] = obj
            else:
                data['obj'] = obj
        rest_request = requests.request(method, rest_service,
                                        params=params, data=data,
                                        headers={'Authorization': authorization})
        status_code = rest_request.status_code
        if status_code == requests.codes.ok:  # pylint: disable=no-member
            claims = rest_request.json()
            if tokens_cache is not None:
                tokens_cache.set_value(cache_key, claims)
            return status_code, claims
        return status_code, {}

    def get_tokens(self, request, credentials):  # pylint: disable=unused-argument
        """Get new tokens from authentication authority"""
        configuration = IJWTSecurityConfiguration(self.context)
        method, service = configuration.get_token_service  # pylint: disable=unpacking-non-sequence
        rest_service = '{}{}'.format(configuration.authority, service)
        if method == 'GET':
            rest_request = requests.request(method, rest_service,
                                            params=credentials.attributes,
                                            allow_redirects=False,
                                            verify=configuration.verify_ssl)
        else:
            rest_request = requests.request(method, rest_service,
                                            data=credentials.attributes,
                                            allow_redirects=False,
                                            verify=configuration.verify_ssl)
        status_code = rest_request.status_code
        if status_code == requests.codes.ok:  # pylint: disable=no-member
            result = rest_request.json()
            result[configuration.access_token_name] = \
                result.pop(configuration.proxy_access_token_name)
            result[configuration.refresh_token_name] = \
                result.pop(configuration.proxy_refresh_token_name)
            return status_code, result
        return status_code, {}

    def refresh_token(self, request):
        """Get new access token with refresh token authorization"""
        configuration = IJWTSecurityConfiguration(self.context)
        method, service = configuration.refresh_token_service  # pylint: disable=unpacking-non-sequence
        rest_service = '{}{}'.format(configuration.authority, service)
        rest_request = requests.request(method, rest_service, headers={
            'Authorization': request.headers.get('Authorization')
        })
        status_code = rest_request.status_code
        if status_code == requests.codes.ok:  # pylint: disable=no-member
            result = rest_request.json()
            result[configuration.access_token_name] = \
                result.pop(configuration.proxy_access_token_name)
            return status_code, result
        return status_code, {}



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""
This module contains PyAMS JWT authentication package package
"""
import os
from setuptools import setup, find_packages


DOCS = os.path.join(os.path.dirname(__file__),
                    'docs')

README = os.path.join(DOCS, 'README.rst')
HISTORY = os.path.join(DOCS, 'HISTORY.rst')

version = '2.0.1'
long_description = open(README).read() + '\n\n' + open(HISTORY).read()

tests_require = [
    'pyams_form',
    'pyams_layer',
    'pyams_site',
    'pyams_skin',
    'pyams_viewlet',
    'pyams_zmi',
    'pyramid_zodbconn'
]


setup(name='pyams_auth_jwt',
      version=version,
      description="PyAMS security plug-in for JWT authentication",
      long_description=long_description,
      classifiers=[
          "License :: OSI Approved :: Zope Public License",
          "Development Status :: 4 - Beta",
          "Programming Language :: Python",
          "Framework :: Pyramid",
          "Topic :: Software Development :: Libraries :: Python Modules",
      ],
      keywords='Pyramid PyAMS',
      author='Thierry Florac',
      author_email='tflorac@ulthar.net',
      url='https://pyams.readthedocs.io',
      license='ZPL',
      packages=find_packages('src'),
      package_dir={'': 'src'},
      namespace_packages=[],
      include_package_data=True,
      package_data={'': ['*.zcml', '*.txt', '*.pt', '*.pot', '*.po', '*.mo',
                         '*.png', '*.gif', '*.jpeg', '*.jpg', '*.css', '*.js']},
      zip_safe=False,
      # uncomment this to be able to run tests with setup.py
      test_suite="pyams_auth_jwt.tests.test_utilsdocs.test_suite",
      tests_require=tests_require,
      extras_require=dict(test=tests_require),
      install_requires=[
          'setuptools',
          # -*- Extra requirements: -*-
          'ZODB',
          'cornice',
          'persistent',
          'pyams_security',
          'pyams_utils >= 1.3.0',
          'pyjwt[crypto]',
          'pyramid >= 2.0.0',
          'requests',
          'zope.container',
          'zope.interface',
          'zope.schema'
      ],
      entry_points="")



#
# Copyright (c) 2015-2019 Thierry Florac <tflorac AT ulthar.net>
# All Rights Reserved.
#
# This software is subject to the provisions of the Zope Public License,
# Version 2.1 (ZPL).  A copy of the ZPL should accompany this distribution.
# THIS SOFTWARE IS PROVIDED "AS IS" AND ANY AND ALL EXPRESS OR IMPLIED
# WARRANTIES ARE DISCLAIMED, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
# WARRANTIES OF TITLE, MERCHANTABILITY, AGAINST INFRINGEMENT, AND FITNESS
# FOR A PARTICULAR PURPOSE.
#

"""PyAMS JWT authentication package.interfaces module

This module provides public interfaces.
"""

from zope.interface import Attribute, Interface, Invalid, invariant
from zope.schema import Bool, Choice, Int, Text, TextLine

from pyams_security.interfaces.plugin import IAuthenticationPlugin
from pyams_utils.cache import BEAKER_CACHES_VOCABULARY
from pyams_utils.schema import HTTPMethodField


__docformat__ = 'restructuredtext'

from pyams_auth_jwt import _


#
# API endpoints
#

REST_TOKEN_ROUTE = 'jwt.rest.token'
REST_TOKEN_PATH = '/api/auth/jwt/token'

REST_VERIFY_ROUTE = 'jwt.rest.verify'
REST_VERIFY_PATH = '/api/auth/jwt/verify'


#
# JWT authentication utility interface
#

JWT_CONFIGURATION_KEY = 'pyams_auth_jwt.configuration'
"""Main JWT configuration key"""

ACCESS_OBJECT = 'access'
"""Object of JWT token used for authentication"""

ACCESS_TOKEN_NAME = 'accessToken'
"""Default name of JSON access token attribute"""

REFRESH_OBJECT = 'refresh'
"""Object of JWT token used for refresh"""

REFRESH_TOKEN_NAME = 'refreshToken'
"""Default name of JSON refresh token attribute"""

JWT_PROXY_CACHE_NAME = 'jwt_tokens'
"""Name of the Beaker cache used to store validated tokens"""

JWT_PROXY_TOKENS_NAMESPACE = 'PyAMS-auth-jwt::tokens'
"""Cache namespace used to store validated tokens"""


class IJWTProxyHandler(Interface):
    """JWT proxy handler"""

    def get_claims(self, request, obj=None):
        """Get claims from given request authorization header"""

    def get_tokens(self, request, credentials):
        """Get new tokens from authentication authority"""

    def refresh_token(self, request):
        """Get new access token with refresh token authorization"""


class IJWTSecurityConfiguration(Interface):
    """Security manager configuration interface for JWT"""

    access_token_name = TextLine(title=_("Access token attribute"),
                                 description=_("Name of the JSON attribute containing "
                                               "access token returned by REST APIs"),
                                 required=False,
                                 default=ACCESS_TOKEN_NAME)

    refresh_token_name = TextLine(title=_("Refresh token attribute"),
                                  description=_("Name of the JSON attribute containing "
                                                "refresh token returned by REST APIs"),
                                  required=False,
                                  default=REFRESH_TOKEN_NAME)

    enabled = Attribute("Enabled configuration?")

    local_mode = Bool(title=_("Enable JWT direct authentication?"),
                      description=_("Enable direct login via JWT authentication"),
                      required=False,
                      default=False)

    algorithm = Choice(title=_("JWT encoding algorithm"),
                       description=_("HS* protocols are using the secret, while RS* protocols "
                                     "are using RSA keys"),
                       required=False,
                       values=('RS256', 'RS512', 'HS256', 'HS512'),
                       default='RS512')

    secret = TextLine(title=_("JWT secret"),
                      description=_("This secret is required when using HS* encryption"),
                      required=False)

    private_key = Text(title=_("JWT private key"),
                       description=_("The secret key is required when using RS* algorithm"),
                       required=False)

    public_key = Text(title=_("JWT public key"),
                      description=_("The public key is required when using RS* algorithm"),
                      required=False)

    access_expiration = Int(title=_("Access token lifetime"),
                            description=_("JWT access token lifetime, in seconds"),
                            required=False,
                            default=60 * 60)

    refresh_expiration = Int(title=_("Refresh token lifetime"),
                             description=_("JWT refresh token lifetime, in seconds"),
                             required=False,
                             default=60 * 60 * 24 * 7)

    proxy_mode = Bool(title=_("Enable JWT proxy authentication?"),
                      description=_("If this option is enabled, tokens management requests "
                                    "will be forwarded to another authentication authority"),
                      required=False,
                      default=False)

    authority = TextLine(title=_("Authentication authority"),
                         description=_("Base URL (protocol and hostname) of the authentication "
                                       "authority to which tokens management requests will be "
                                       "forwarded"),
                         required=False)

    get_token_service = HTTPMethodField(title=_("Token getter service"),
                                        description=_("REST HTTP service used to get a new token"),
                                        required=False,
                                        default=('POST', REST_TOKEN_PATH))

    proxy_access_token_name = TextLine(title=_("Access token attribute"),
                                       description=_("Name of the JSON attribute returned by "
                                                     "REST API containing access tokens"),
                                       required=False,
                                       default=ACCESS_TOKEN_NAME)

    proxy_refresh_token_name = TextLine(title=_("Refresh token attribute"),
                                        description=_("Name of the JSON attribute returned by "
                                                      "REST API containing refresh tokens"),
                                        required=False,
                                        default=REFRESH_TOKEN_NAME)

    get_claims_service = HTTPMethodField(title=_("Token claims getter"),
                                         description=_("REST HTTP service used to extract claims "
                                                       "from provided authorization token"),
                                         required=False,
                                         default=('GET', REST_TOKEN_PATH))

    refresh_token_service = HTTPMethodField(title=_("Token refresh service"),
                                            description=_("REST HTTP service used to get a new "
                                                          "access token with a refresh token"),
                                            required=False,
                                            default=('PATCH', REST_TOKEN_PATH))

    verify_token_service = HTTPMethodField(title=_("Token verify service"),
                                           description=_("REST HTTP service used to check "
                                                         "validity of an existing token"),
                                           required=False,
                                           default=('POST', REST_VERIFY_PATH))

    verify_ssl = Bool(title=_("Verify SSL?"),
                      description=_("If 'no', SSL certificates will not be verified"),
                      required=False,
                      default=True)

    use_cache = Bool(title=_("Use verified tokens cache?"),
                     description=_("If selected, this option allows to store credentials in a "
                                   "local cache from which they can be reused"),
                     required=False,
                     default=True)

    selected_cache = Choice(title=_("Selected tokens cache"),
                            description=_("Beaker cache selected to store validated tokens"),
                            required=False,
                            vocabulary=BEAKER_CACHES_VOCABULARY,
                            default='default')

    @invariant
    def check_configuration(self):
        """Check for JWT configuration"""
        if self.local_mode and self.proxy_mode:
            raise Invalid(_("You can't enable both local and proxy modes"))
        if self.local_mode:
            if not self.algorithm:
                raise Invalid(_("You must choose an algorithm to enable JWT authentication"))
            if self.algorithm.startswith('HS'):  # pylint: disable=no-member
                if not self.secret:
                    raise Invalid(_("You must define JWT secret to use HS256 algorithm"))
            elif self.algorithm.startswith('RS'):  # pylint: disable=no-member
                if not (self.private_key and self.public_key):
                    raise Invalid(_("You must define a private and a public key to use "
                                    "RS256 algorithm"))
        if self.proxy_mode:
            if not self.authority:
                raise Invalid(_("You must define authentication authority to use proxy mode"))
            if self.use_cache and not self.selected_cache:
                raise Invalid(_("You must choose a cache to enable tokens caching"))


class IJWTAuthenticationPlugin(IAuthenticationPlugin):
    """JWT authentication plugin"""

    configuration = Attribute("JWT configuration")
    enabled = Attribute("Enable JWT authentication?")

    audience = Attribute("Token audience")
    leeway = Attribute("Token leeway")
    http_header = Attribute("HTTP header used for JWT token")
    auth_type = Attribute("JWT authentication type")
    callback = Attribute("JWT authentication callback")
    json_encoder = Attribute("JSON encoder used to encode token claims")

    def create_token(self, principal, expiration=None, audience=None, **claims):
        """Create JWT token"""

    def get_claims(self, request):
        """Extract claims from JWT token"""
